;=======================================================================
TITLE	KEYMAP - Remap Keyboard on a PC
;=======================================================================
;   SOURCE : Keymap.asm
;   DATE   : Oct. 10, 2001
;   VERSION: 1.00
;
;   Public Domain code ( originally by Alan D. Jones )
;
;   Modified: M. Genovy
;
;=======================================================================
;
;   Program Description:
;   --------------------
;
;   This program  remaps the keyboard on an AT pc. BIOS interrupt 15
;   is used to do this. This program intercepts BIOS Interrupt 15 calls
;   and translates the scan codes for the keyboard. The translated scan
;   codes are passed on to the normal Interrupt 15 handler.
;
;
;-----------------------------------------------------------------------
;
; TASM keymap.asm /la
; TLINK keymap
;

.MODEL Small                   ; Use small model for a .COM program

.CODE

assume cs:_TEXT,ds:_TEXT

org 100h

Start:
       Jmp       load          ; Jump to load the code
;
;----------------------------------------------------------------------
;                  Keyboard Make-Break Scan Codes
;                  ------------------------------
; From Programming Solutions Handboook for Microcomputers
; by  Julio Sanchez & Maria P. Canton   Page 239
;----------------------------------------------------------------------
;  KEY      AT PC          KEY         AT PC       KEY         AT PC
;
; <ESC>    01H-81h     <Scroll Lock>  46h-C6h   <Up Arrow>    48h-C8h
; <F1>     3Bh-BBh     <Back Space>   0Eh-8Eh   <Left Arrow>  4Bh-CBh
; <F2>     3Ch-BCh     <Ins>          52h-D2h   <Right Arrow> 4Ch-CCh
; <F3>     3Dh-BDh     <Del>          53h-D3h   <Down Arrow>  50h-D0h
; <F4>     3Eh-BEh     <Num Lock>     45h-C5h
; <F5>     3Fh-BFh     <Tab>          0Fh-8Fh
; <F6>     40h-C0h     <Enter>        1Ch-9Ch
; <F7>     41h-C1h     <Caps Lock>    3Ah-BAh
; <F8>     42h-C2h     <Left Shift>   2Ah-AAh
; <F9>     43h-C3h     <Right Shift>  36h-B6h
; <F10>    44h-C4h     <Left Ctrl>    1Dh-9Dh
;
;----------------------------------------------------------------------
;                     Scan Code Translation Table
;----------------------------------------------------------------------
; This is a complete scan code translation table. The scan code
; retrieved from the keyboard is used as an index into this table, and
; the byte found at "scantbl+<scancode>" is substituted and passed
; to the keyboard interrupt handler.
;
; CAUTION: If this table is modified, be sure to swap or substitute
;          matching pairs ! Both the "make" code and the "break" code
;          ( = make code + 80h ) must be changed.
;
; <left arrow>  key: make = 4Bh, break = 0CBh
; <F3>          key: make = 3Dh, break = 0BDh
;
; <right arrow> key: make = 4Dh, break = 0CDh
; <F1>          key: make = 3Bh, break = 0BBh
;
; The table as it stands exchanges the <left arrow> and <F3> key
; and the <right arrow> and <F1> key.
;----------------------------------------------------------------------
;
org 104h

scantbl DB 000h,001h,002h,003h,004h,005h,006h,007h
        DB 008h,009h,00Ah,00Bh,00Ch,00Dh,00Eh,00Fh
        DB 010h,011h,012h,013h,014h,015h,016h,017h
        DB 018h,019h,01Ah,01Bh,01Ch,01Dh,01Eh,01Fh
        DB 020h,021h,022h,023h,024h,025h,026h,027h
        DB 028h,029h,02Ah,02Bh,02Ch,02Dh,02Eh,02Fh
        DB 030h,031h,032h,033h,034h,035h,036h,037h
        DB 038h,039h,03Ah,04Dh,03Ch,04Bh,03Eh,03Fh
        DB 040h,041h,042h,043h,044h,045h,046h,047h
        DB 048h,049h,04Ah,03Dh,04Ch,03Bh,04Eh,04Fh
        DB 050h,051h,052h,053h,054h,055h,056h,057h
        DB 058h,059h,05Ah,05Bh,05Ch,05Dh,05Eh,05Fh
        DB 060h,061h,062h,063h,064h,065h,066h,067h
        DB 068h,069h,06Ah,06Bh,06Ch,06Dh,06Eh,06Fh
        DB 070h,071h,072h,073h,074h,075h,076h,077h
        DB 078h,079h,07Ah,07Bh,07Ch,07Dh,07Eh,07Fh
        DB 080h,081h,082h,083h,084h,085h,086h,087h
        DB 088h,089h,08Ah,08Bh,08Ch,08Dh,08Eh,08Fh
        DB 090h,091h,092h,093h,094h,095h,096h,097h
        DB 098h,099h,09Ah,09Bh,09Ch,09Dh,09Eh,09Fh
        DB 0A0h,0A1h,0A2h,0A3h,0A4h,0A5h,0A6h,0A7h
        DB 0A8h,0A9h,0AAh,0ABh,0ACh,0ADh,0AEh,0AFh
        DB 0B0h,0B1h,0B2h,0B3h,0B4h,0B5h,0B6h,0B7h
        DB 0B8h,0B9h,0BAh,0CDh,0BCh,0CBh,0BEh,0BFh
        DB 0C0h,0C1h,0C2h,0C3h,0C4h,0C5h,0C6h,0C7h
        DB 0C8h,0C9h,0CAh,0BDh,0CCh,0BBh,0CEh,0CFh
        DB 0D0h,0D1h,0D2h,0D3h,0D4h,0D5h,0D6h,0D7h
        DB 0D8h,0D9h,0DAh,0DBh,0DCh,0DDh,0DEh,0DFh
        DB 0E0h,0E1h,0E2h,0E3h,0E4h,0E5h,0E6h,0E7h
        DB 0E8h,0E9h,0EAh,0EBh,0ECh,0EDh,0EEh,0EFh
        DB 0F0h,0F1h,0F2h,0F3h,0F4h,0F5h,0F6h,0F7h
        DB 0F8h,0F9h,0FAh,0FBh,0FCh,0FDh,0FEh,0FFh
;
; The table below is an UNMODIFIED table for reference
;
;scantbl DB 000h,001h,002h,003h,004h,005h,006h,007h
;        DB 008h,009h,00Ah,00Bh,00Ch,00Dh,00Eh,00Fh
;        DB 010h,011h,012h,013h,014h,015h,016h,017h
;        DB 018h,019h,01Ah,01Bh,01Ch,01Dh,01Eh,01Fh
;        DB 020h,021h,022h,023h,024h,025h,026h,027h
;        DB 028h,029h,02Ah,02Bh,02Ch,02Dh,02Eh,02Fh
;        DB 030h,031h,032h,033h,034h,035h,036h,037h
;        DB 038h,039h,03Ah,03Bh,03Ch,03Dh,03Eh,03Fh
;        DB 040h,041h,042h,043h,044h,045h,046h,047h
;        DB 048h,049h,04Ah,04Bh,04Ch,04Dh,04Eh,04Fh
;        DB 050h,051h,052h,053h,054h,055h,056h,057h
;        DB 058h,059h,05Ah,05Bh,05Ch,05Dh,05Eh,05Fh
;        DB 060h,061h,062h,063h,064h,065h,066h,067h
;        DB 068h,069h,06Ah,06Bh,06Ch,06Dh,06Eh,06Fh
;        DB 070h,071h,072h,073h,074h,075h,076h,077h
;        DB 078h,079h,07Ah,07Bh,07Ch,07Dh,07Eh,07Fh
;        DB 080h,081h,082h,083h,084h,085h,086h,087h
;        DB 088h,089h,08Ah,08Bh,08Ch,08Dh,08Eh,08Fh
;        DB 090h,091h,092h,093h,094h,095h,096h,097h
;        DB 098h,099h,09Ah,09Bh,09Ch,09Dh,09Eh,09Fh
;        DB 0A0h,0A1h,0A2h,0A3h,0A4h,0A5h,0A6h,0A7h
;        DB 0A8h,0A9h,0AAh,0ABh,0ACh,0ADh,0AEh,0AFh
;        DB 0B0h,0B1h,0B2h,0B3h,0B4h,0B5h,0B6h,0B7h
;        DB 0B8h,0B9h,0BAh,0BBh,0BCh,0BDh,0BEh,0BFh
;        DB 0C0h,0C1h,0C2h,0C3h,0C4h,0C5h,0C6h,0C7h
;        DB 0C8h,0C9h,0CAh,0CBh,0CCh,0CDh,0CEh,0CFh
;        DB 0D0h,0D1h,0D2h,0D3h,0D4h,0D5h,0D6h,0D7h
;        DB 0D8h,0D9h,0DAh,0DBh,0DCh,0DDh,0DEh,0DFh
;        DB 0E0h,0E1h,0E2h,0E3h,0E4h,0E5h,0E6h,0E7h
;        DB 0E8h,0E9h,0EAh,0EBh,0ECh,0EDh,0EEh,0EFh
;        DB 0F0h,0F1h,0F2h,0F3h,0F4h,0F5h,0F6h,0F7h
;        DB 0F8h,0F9h,0FAh,0FBh,0FCh,0FDh,0FEh,0FFh
;

old15   dd    0                 ;

;----------------------------------------------------------------------
;                       New Interrupt 15 Routine
;----------------------------------------------------------------------
; Interrupt 15h Function 4Fh is called ( on most AT BIOS's ) by the
; keyboard interrupt handler, with the Scan Code passed in the
; AL register. This program traps the Int15h Function 4Fh calls and
; examines the Scan Code BEFORE it is processed by the handler. If
; Int15h returns with the Carry set, then processing proceeds. If it
; returns with the Carry clear, then processing is skipped, and the
; Bios ignores the keypress.
;
;  ie. Int09h calls Int15h, and if Int15h returns with carry set
;      then Int09h will process the key press
;

int15:
      pushf                     ;
      cmp       ah,4fh          ; See if this is scan code call
      je        scancode        ;
;
;----------------------------------
; No - act as if we are not here
;----------------------------------
;
      popf                      ;
      jmp       cs:old15        ; Do the old interrupt 15 code
;
;----------------------------------------------------------------------
; Yes - so run the scan code through the translator
;----------------------------------------------------------------------
;
scancode:

      popf                      ;
      push      ds              ;
      push      bx              ;
      push      cs              ;
      pop       ds              ;
;
; AL at this point contains the Scan Code. BX is the Base address of the
; translate table. Xlatb loads AL with the byte pointed to by BX:AL
;
      lea       bx,scantbl      ; Get translated Scan Code
      xlatb                     ; Translate (AL is index, BX is base address)
      pop       bx              ;
      pop       ds              ;
;
; Now set the Carry flag so that the Scan Code is processed by Int09h
;
      stc                       ;
      iret                      ;
;
;----------------------------------------------------------------------
;                    Initialization Routine
;----------------------------------------------------------------------
; All code beyond this point is not preserved after exit
;
cutoff  label   byte
;
; Message if PC not an AT type
;
msg1    db      'This program only runs '
        db      ' on PC/AT type machines '
        db      0Dh,0Ah,24h

load:
;
; Make sure this is a PC/AT machine
;
      push      sp                         ; Test 1 -Flag Register Test
      pop       ax                         ; bits 12 to 15 set to 1's on
      cmp       ax,sp                      ; an XT, but not on AT's
      jne       wrongmachine               ;
;
      mov       ax,0F000h                  ; Test 2 - Machine ID Byte
      mov       es,ax                      ;
      cmp       byte ptr es:[0FFFEh],0FCh  ;  if its 0FCh then its AT !
      jne       wrongmachine               ;
;
; Get the old interrupt vector
;
      mov       ax,3515h            ; Get Interrupt Vector for Int15h
      int       21h                 ;
      mov       word ptr old15,bx   ; and save it
      mov       word ptr old15+2,es ;
;
; Install new Interrupt 15 Routine
;
       mov      dx,offset int15 ; Get address of new Int15h routine
       mov      ax,2515h        ; Set new Interrupt Vector for Interrupt 15
       int      21h             ;
;
; Now exit and leave new Interrupt 15 code resident
;
       mov      ax,3100h        ; Code for exit and remain resident (TSR )
       lea      dx,cutoff       ; Length of interrupt routine
       add      dx,0Fh          ; Convert to number of segments
       shr      dx,1            ;
       shr      dx,1            ;
       shr      dx,1            ;
       shr      dx,1            ;
       int      21h             ; Exit and remain resident
;
; Display Wrong machine error message and exit
;
wrongmachine:

       lea      dx,msg1         ; Pointer to message
       mov      ah,9            ; DOS function - Display String
       int      21h             ;
       mov      ax,4C01h        ; DOS function - Terminate with
       int      21h             ;                Return Code

       end      start           ;
